/*	swiper.c - M.A.R-T>Y. RFID Tagger and Bagger daemon written in C
	Copyright (C) 2023-2024 hoff.industries

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

#include <poll.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>

#include <errno.h>
#include <limits.h>

#ifndef LIBMMP_AUTHOR
	#include "libmmp.h"
#endif

#ifndef LIBMARTY_AUTHOR
	#include "libmarty.h"
#endif

#include "swiperconf.h"

#ifndef MENDPOINTS_H
	#include "mendpoints.h"
#endif

#define CODE_BUFFER_LEN 256

const int TARGET_PROTOCOL[3] = {2, 0, 0};
const int INSTANCE_VERSION[3] = {3, 0, 2};

const char INSTANCE_NAME[] = "m-cswiper";
const char INSTANCE_TITLE[] = "Swiper - RFID Tagger and Bagger";

struct MartyClientEndpoint Endpoints[] = {
	{
		.endpoint = SWIPER_EPT_CYCLE,
		.name = "cycle",
		.min = 7,
		.max = 7,
		.isAscii = true
	}
};

int MaxEndpoint = sizeof(Endpoints) / sizeof(Endpoints[0]);

void *application_loop(void *voidData) {

	struct MartyClientData *data = (struct MartyClientData *)voidData;

	char codeBuffer[CODE_BUFFER_LEN] = {0};
	char *readLine = NULL;

	int eventIndex = -1;

	struct MMStatus status;
	struct RFIDEvent *event = NULL;

	if (data == NULL) {
		marty_part_log("Void pointer passed to application loop.", LOG_CRITICAL);
		exit(EXIT_ERR_BAD_POINTER);
	}

	mmstatus_factory(&status);

	while (true) {

		errno = 0;
		readLine = fgets(codeBuffer, CODE_BUFFER_LEN, stdin);

		if (readLine == NULL) {

			if (errno == 0) {
				continue;
			}

			marty_full_log("Error occurred reading from stdin!\n", LOG_ERROR, data);
			exit(EXIT_ERR_READ);
		}

		eventIndex = strlen(readLine);
		if (readLine[eventIndex - 1] == '\n') {

			// get rid of newline
			readLine[eventIndex - 1] = '\0';
		}

		eventIndex = get_rfid_event(readLine);
		if (eventIndex == -1) {

			marty_full_log("Event code not found in register!\n", LOG_WARNING, data);
			continue;
		}

		event = &EventRegister[eventIndex];

		if (event->current != NULL) {
			event = event->current;
		}

		if (event->handle == NULL) {

			marty_full_log("NULL handle found in register!\n", LOG_WARNING, data);
			goto JMP_INCR_LINKED_LIST;
		}

		marty_message_handler(event->handle, &status, data);
		if (status.error != MMP_ERR_NONE) {

			marty_full_log("Event execution failure!\n", LOG_ERROR, data);
		}

		JMP_INCR_LINKED_LIST:

		EventRegister[eventIndex].current = event->next;
	}
}

int main(int argc, char **argv) {
	pthread_t networkThread, applicationThread;
	struct MartyClientData data;
	struct MMStatus status;

	int bytesRead = 0;
	struct MMPacket txPacket;
	struct MMPacket *rxPacket = NULL;

	mmp_arg applicationStatus = MMP_APPLICATION_SUCCESS;

	char hostname[_POSIX_HOST_NAME_MAX];

	int endpointConversion = -1;
	struct MartyClientEndpoint *endpoint = NULL;

	struct MartyMessageHandle *deliveryHandle = NULL;

	int eventIndex;

	printf("%s[%s]%s Starting hoff.industries %s - %s v%d.%d.%d\n",
		   LoggerColours[LOG_INFO], MartyFlavour, ClearColour, INSTANCE_TITLE,
		INSTANCE_NAME, INSTANCE_VERSION[0], INSTANCE_VERSION[1], INSTANCE_VERSION[2]);

	if (marty_client_data_factory(&data) == -1) {
		marty_part_log("Client data init failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_CDATA);
	}

	mmstatus_factory(&status);
	mmpacket_factory(&txPacket, true);

	data.versionString = alloc_marty_version_string(INSTANCE_VERSION, INSTANCE_NAME);
	if (data.versionString == NULL) {
		marty_part_log("Version string allocation failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_VSTRING);
	}

	if (gethostname(hostname, _POSIX_HOST_NAME_MAX) == -1) {
		marty_part_log("Hostname get failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_HOSTNAME);
	}

	marty_part_log("Loaded ", LOG_STARTUP);
	printf("%d events from the register.\n", EventLength);

	data.connType = CONN_UNIX;
	data.initParameters.auth = "";
	data.initParameters.priority = 0;
	data.initParameters.host = hostname;
	data.initParameters.name = "swiper";

	if (spawn_marty_socket(&data) != MMP_ERR_NONE) {
		marty_full_log("Socket spawn failure!\n", LOG_CRITICAL, &data);
		exit(EXIT_ERR_SPAWN);
	}

	if (marty_init(&data) != MMP_ERR_NONE) {
		marty_full_log("MMP initialisation failure!\n", LOG_CRITICAL, &data);
		exit(EXIT_ERR_INIT);
	};

	marty_full_log("MMP initialisation success!\n", LOG_STARTUP, &data);

	if (pthread_create(&networkThread, NULL, marty_network_loop, (void *)&data) != 0) {

		marty_full_log("Error spawning network thread!\n", LOG_CRITICAL, &data);
		exit(EXIT_ERR_SPAWN);
	}
	marty_full_log("Network loop started.\n", LOG_STARTUP, &data);

	if (pthread_create(&applicationThread, NULL, application_loop, (void *)&data) != 0) {

		marty_full_log("Error spawning application thread!\n", LOG_CRITICAL, &data);
		exit(EXIT_ERR_SPAWN);
	}
	marty_full_log("Application loop started.\n", LOG_STARTUP, &data);

	while (true) {
		mmpacket_reset(&txPacket, true);
		applicationStatus = MMP_APPLICATION_SUCCESS;

		if (rxPacket != NULL) {
			mmpacket_reset(rxPacket, true);
			free(rxPacket);
		}

		rxPacket = NULL;
		deliveryHandle = NULL;

		if (poll(data.pfds, data.nfds, -1) == -1) {
			marty_full_log("Error polling inter-thread pipes!\n", LOG_CRITICAL, &data);
			exit(EXIT_ERR_POLL);
		}

		if (data.pfds[ITC_DELIVERY_PIPE].revents != 0) {
			// check APPLICATION events

			bytesRead = read(data.netDeliveryPipe[0], &deliveryHandle,
				sizeof(struct MartyMessageHandle *));

			if (bytesRead == 0 || bytesRead == -1) {
				marty_full_log("Error reading inter-thread pipes!\n", LOG_CRITICAL, &data);
				exit(EXIT_ERR_ITC);
			}

			if (deliveryHandle == NULL) {
				// if ITC pipe sent -2 means reconnect event occurred.
				continue;
			}

			switch (deliveryHandle->result) {
				case MMP_APPLICATION_SUCCESS:
					break;

				case MMP_APPLICATION_DELIVERY:

					marty_full_log("Failed to execute event: delivery failure. Ohhh mayyynnn!\n", LOG_ERROR, &data);
					continue;

				default:
					continue;
			}

			if (deliveryHandle->next == NULL) {
				marty_full_log("Sequence execution success. Pipe pipe pipe, piper I'm pipin'\n", LOG_SUCCESS, &data);
				continue;
			}

			marty_message_handler(deliveryHandle->next, &status, &data);
			if (status.error != MMP_ERR_NONE) {
				marty_full_log("Error executing follow-on sequence member!\n", LOG_WARNING, &data);
			}

			continue;
		}

		if (data.pfds[ITC_MESSAGE_PIPE].revents == 0) {
			// since MESSAGE events are last on our list to check
			// we can continue if we don't need to handle
			continue;
		}

		bytesRead = read(data.netMessagePipe[0], &rxPacket,
			sizeof(struct MMPacket *));

		if (bytesRead == 0 || bytesRead == -1) {
			marty_full_log("Error reading inter-thread pipes!\n", LOG_CRITICAL, &data);
			exit(EXIT_ERR_ITC);
		}

		if (rxPacket == NULL) {
			marty_full_log("Received NULL pointer from ITC pipe!\n", LOG_WARNING, &data);
			continue;
		}

		endpointConversion = rxPacket->payloads[PAYL_IDX_ENDPOINT][0];

		if (endpointConversion > MaxEndpoint || endpointConversion < 0) {
			marty_full_log("Requested endpoint outside my range!\n", LOG_WARNING, &data);
			applicationStatus = MMP_APPLICATION_ENDPOINT;

			goto JMP_TX_STATUS;
		}

		endpoint = &Endpoints[endpointConversion];

		if (rxPacket->payloadLen < endpoint->min) {
			marty_full_log("Request payload count less than expected!\n", LOG_WARNING, &data);
			applicationStatus = MMP_APPLICATION_TOO_FEW;

			goto JMP_TX_STATUS;
		}

		if (rxPacket->payloadLen > endpoint->max) {
			marty_full_log("Request payload count more than expected!\n", LOG_WARNING, &data);
			applicationStatus = MMP_APPLICATION_TOO_MANY;

			goto JMP_TX_STATUS;
		}

		switch (endpointConversion) {
			case SWIPER_EPT_CYCLE:

				// to be honest I don't really care about the
				// possible race condition here; timing would
				// need to be insane and it would just exec
				// the wrong event once... whoop de doo

				rxPacket->payloads[5][mmpayloadlen(rxPacket->payloads[5])] = '\0';
				if (!strcmp((char *)rxPacket->payloads[5], "all")) {

					for (int i = 0; i < EventLength; i++) {
						EventRegister[i].current = NULL;
					}

					goto JMP_TX_STATUS;
				}

				eventIndex = get_rfid_event((char *)rxPacket->payloads[5]);
				if (eventIndex == -1) {

					marty_full_log("Requested reset code not in register!\n", LOG_WARNING, &data);
					applicationStatus = MMP_APPLICATION_EXECUTE;

					goto JMP_TX_STATUS;
				}

				EventRegister[eventIndex].current = NULL;

				break;

			default:
				marty_full_log("Not handling this endpoint.\n", LOG_WARNING, &data);
				applicationStatus = MMP_APPLICATION_ENDPOINT;

				goto JMP_TX_STATUS;
		}

		JMP_TX_STATUS:

		marty_set_application(&txPacket, &status, rxPacket->payloads[0], applicationStatus);
		if (status.error != MMP_ERR_NONE) {
			marty_full_log("Couldn't set autopacket!\n", LOG_ERROR, &data);
			continue;
		}

		mmp_serialize(&txPacket, &status);
		if (status.error != MMP_ERR_NONE) {
			marty_full_log("Couldn't serialize autopacket!\n", LOG_ERROR, &data);
			continue;
		}

		marty_net_write(&txPacket, &status, &data);
	}
}
